"""
    Action update
"""

from lib.actions import Action
from lib.action import actionrm
from lib.alpm import AlpmBranch
from lib.exceptions import BranchNotFoundError


class ActionUpdate(Action):

    @staticmethod
    def usage():
        return """
            update [--force or -f] [--files] [--mirrors | -m] : update all branches
            update --testing [--force or -f] [--files]: update one branch
            """

    def run(self, branch_name: str = ''):
        """ create or update databases """
        super().run()
        from_gui = self.params.option('gui')
        self.force = self.params.option('force', 'f')
        files = self.params.option('files')

        self.pacman_mirror()

        def update_branch(branch_name: str, from_gui: bool):
            branch = AlpmBranch(branch_name)
            if from_gui:
                print(f":: Branch: {branch_name}")
                branch.do_update(self.force)
            else:
                self.msg(f"Branch: {branch.color(branch_name)}{branch_name}{branch.color('none')}")
                branch.do_refresh(self.force, files)

        if branch_name:
            self.params.branch = branch_name
        try:
            if self.params.branch:
                update_branch(self.params.branch, from_gui)
            else:
                for br in AlpmBranch.BRANCHES:
                    update_branch(br, from_gui)
        except BranchNotFoundError:
            # structure not exists -> create folders and .conf
            for br in AlpmBranch.BRANCHES:
                AlpmBranch.create_branch(br)
                self.msg(f"Branch: {AlpmBranch.color(br)}{br}{AlpmBranch.color('none')}")
                if from_gui:
                    AlpmBranch(br).do_update(self.force)
                else:
                    AlpmBranch(br).do_refresh(True, files)
        else:
            raise

    def pacman_mirror(self):
        if not self.params.option('mirrors', 'm') or self.params.option('gui'):
            return

        import urllib, json

        with urllib.request.urlopen('https://repo.manjaro.org/status.json') as f_url:
            req = f_url.read()#.decode('utf-8')
        mirrors = json.loads(req)
        mirrors = [m for m in mirrors if m['branches'] == [1, 1, 1]]
        sorted(mirrors, key=lambda colonnes: colonnes['country'])
        i = 0
        for i, mirror in enumerate(mirrors):
            print(f"{i+1:2}", f"{mirror['country']:20} {mirror['last_sync']:7} {mirror['url']}")
        id_choice = -1
        i += 1

        while id_choice < 0:
            value = input(f"Enter a number (1..{i}) : ")
            if not value:
                break
            try:
                if int(value) <= i:
                    id_choice = int(value)
            except ValueError:
                pass

        if id_choice > 0:
            id_choice -= 1
            print(f"\n   {mirrors[id_choice]['country']:20} {mirrors[id_choice]['last_sync']:7} {mirrors[id_choice]['url']}")
            # update .conf
            self.force = AlpmBranch.set_server(mirrors[id_choice]['url'])
            actionrm.ActionRm(self.params).run()     # remove all cache

        if not self.params.option('files'):
            print("\n:: !! files not updated !!\n")
        #if force:
        #    files = True



#!/usr/bin/python
"""
 ALPM script

 use one pacman.conf by branch
 pacman -Si linux --config /var/lib/manjaro-repos/archlinux/etc/pacman.conf
 pacman -Si linux50 --config /var/lib/manjaro-repos/testing/etc/pacman.conf
 expac -Ss "%n %v"  --config /var/lib/manjaro-repos/testing/etc/pacman.conf

 #security
 pacman -S linux --config /var/lib/manjaro-repos/archlinux/etc/pacman.conf --logfile /var/lib/manjaro-repos/archlinux --dbonly
 Error HOOK and if not, after: filesystem conflict file

"""

import os
import lib
from lib.parser import ConsoleParser
from lib.actions import SwitchClass
from lib import action


################################################

class MainApp:
    """Main Application parse console arguments"""

    def __init__(self, args):
        """ object initialisation """
        if not os.path.exists(lib.TMP_DIR):
            os.mkdir(lib.TMP_DIR)
        self.params = ConsoleParser(args)
        self.actions = SwitchClass(action.ACTIONS)
        self.action = self.actions.factory(self.params.action, self.params)

    def run(self):
        """ run one action """
        if self.action.test():
            self.action.run()
        else:
            print(self.action.error, self.action.usage().replace(' '*11, ''))



"""
        Action list packagers
"""
import os
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionList(Action):

    @staticmethod
    def usage():
        return "list [--unstable] : packagers listing"

    def run(self):
        """manjaro packagers listing"""
        super().run()
        if not self.params.branch:
            self.params.branch = 'unstable'

        branch = AlpmBranch(self.params.branch)
        self.msg(f"Branch: {branch.color(self.params.branch)}{self.params.branch}{branch.color('none')}")
        user = os.environ.get('USER', '????')
        for packager in sorted(set(pkg.packager for pkg in branch.pkg_search('') if 'manjaro' in pkg.packager)):
            items = packager.split('<')
            try:
                ico = ''
                if user in items[1]:
                    ico = '🤘'
                print(f"{items[0]:28}{items[1].rstrip('>'):32} {ico}")
            except IndexError:
                print(packager)



"""
        Action remove
"""
import os
import subprocess
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionRm(Action):

    @staticmethod
    def usage():
        return "rm : remove all local databases"

    def test(self):
        super().test()
        if os.geteuid() != 0:
            raise PermissionError("use sudo for remove local databases")
        return True

    def run(self):
        """remove local databases"""
        super().run()
        if AlpmBranch.ROOT:
            subprocess.check_call(['rm', '-rf', AlpmBranch.ROOT])
            print(f"{AlpmBranch.ROOT} removed")



"""
 format output
 use template and color code
    * colors: ($a,$u,$t,$s,$n) # and $c for curent ?
    * {%n %v:48} {%m:>12}
    * {$a%n$n %v:48} {%m:>12} # with color
"""

import time
from lib.alpm import AlpmBranch

class DisplayTemplate():
    """ general template """

    colors = {
        'a': AlpmBranch.COLORS['archlinux'],
        'u': AlpmBranch.COLORS['unstable'],
        't': AlpmBranch.COLORS['testing'],
        'g': AlpmBranch.COLORS['stable-staging'],
        's': AlpmBranch.COLORS['stable'],
        'n': AlpmBranch.COLORS['none'],
        'i': AlpmBranch.COLORS['i']
    }

    def __init__(self, template: str, color_default: str = 'n'):
        self.template = template
        self.templates = []
        self.results = []
        try:
            if color_default == "stable-staging":
                color_default = "g"
            self.colors['d'] = self.colors.get(color_default[:1], AlpmBranch.COLORS['none'])
        except IndexError:
            self.colors['d'] = AlpmBranch.COLORS['none']
        self.parse()

    def parse(self):
        templates = self.template.split('}')
        self.results = []

        for template in templates:
            if not template:
                continue
            align_left = True
            count = 0
            template = template.replace('{', '')
            pos = template.rfind(":")
            if pos > -1:
                count = template[pos+1:]
                if count.startswith('>'):
                    align_left = False
                    count = count[1:]
                try:
                    count = int(count)
                except ValueError:
                    count = 1
                    pos = 65000
            else:
                pos = 65000
            self.templates.append((template[0:pos], count, align_left)) # template, long

    def push(self, namespace: dict):
        self.results = []
        for template in self.templates:
            result = template[0]
            #push datas
            for key in namespace:
                result = result.replace(f"%{key}", str(namespace.get(key, '?')))
            # align columns
            # count '$' for exclude color codes
            while len(result) - result.count('$')*2 < template[1]:
                if template[2]: # align to left ?
                    result += ' '
                else:
                    result = ' '+result
            #push colors
            # TODO
            # if '$' is in description ?
            for key in self.colors:
                result = result.replace(f"${key}", str(self.colors.get(key, '')))
            self.results.append(result)
        return self.results

    def string(self, namespace=None):
        if namespace:
            self.push(namespace)
        result = ""
        for template in self.results:
            result = result + template
        return result.replace('\\n', "\n").replace('\\t', "\t")

    def debug(self):
        print('format:', self.template)
        for template in self.templates:
            print('template: ', template)
        for result in self.results:
            print('result: ', f"[{result}]")


class DisplayPkg(DisplayTemplate):
    """ template for one pkg """

    def __init__(self, template: str, branchname: str):
        super().__init__(template, branchname)
        self.branch = branchname

    @staticmethod
    def pkg_attr(key):
        """ convert if data is array """
        if key:
            if isinstance(key, list):
                return ' '.join(key)
            else:
                return key
        return ' '

    def push(self, pkg):
        """ push package datas in results """
        namespace = {
            'n': pkg.name,
            'v': pkg.version,
            'p': pkg.packager,
            'P': self.pkg_attr(pkg.provides),
            'D': self.pkg_attr(pkg.depends),
            'u': self.pkg_attr(pkg.url),
            'r': pkg.db.name,
            'd': pkg.desc,
            'G': self.pkg_attr(pkg.groups),
            'b': time.strftime('%Y-%m-%d', time.localtime(pkg.builddate)),
            'e': self.pkg_attr(pkg.base),
            'm': f"{int(pkg.size/1000):>6}", # size Ko
            'w': 'asdep' if pkg.reason == 1 else ''
        }
        return super().push(namespace)



"""
        Action tree
"""
import os
import subprocess
import re
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionTree(Action):

    @staticmethod
    def usage():
        return "tree : local database tree informations"

    def run(self):
        """show local database tree structure"""
        super().run()

        subprocess.call(['tree', '--noreport', '-I', 'local|hooks|pacman.log', AlpmBranch.ROOT])
        regex_kernel = re.compile(r'^linux(\d+)(|-rt)$')
        regex_packager = re.compile(r'manjaro')
        for branchname in AlpmBranch.BRANCHES:
            nbs = {}
            packagers = {}
            linuxs = []
            branch = AlpmBranch(branchname)
            if branchname == 'archlinux':
                continue
                #regex_kernel = re.compile('^(linux|linux-lts)$')
            for db in branch.get_repos():
                pkgs = db.search('')
                nbs[db.name] = len(pkgs)
                if db.name in 'core': #['core', 'community'] add community for to have kernels-rt
                    linuxs = linuxs + [x.name for x in pkgs if regex_kernel.search(x.name)]
                packagers[db.name] = len([x.packager for x in pkgs if regex_packager.search(x.packager)])

            print(f"\n{branch.color(branchname)}:: {branchname:22}{branch.color('none')} {sum(nbs.values()):>6_} {sum(packagers.values()):>10_}".replace('_', ' '))
            for key in nbs:
                print(f"   {branch.color(branchname)}-{branch.color('none')} {key:20} {nbs[key]:>6_} {packagers[key]:>10_}".replace('_', ' '))
            if linuxs:
                print("  ", *reversed(linuxs), sep=" ")

        if not os.path.isfile(f"{AlpmBranch.ROOT}/unstable/etc/pacman.conf"):
            return
        regex_path = re.compile(r'^Server(.?)=(?P<path>[^#]+)')
        with open(f"{AlpmBranch.ROOT}/unstable/etc/pacman.conf", 'r') as conf:
            for line in conf:
                matchs = regex_path.match(line.lstrip())
                if not matchs:
                    continue
                path = matchs.group('path')
                if path:
                    print(f"\nServer: {AlpmBranch.color('i')}{path.strip()[0:-21]}{AlpmBranch.color('none')}")
                    return



#!/usr/bin/python
"""
    gui Gtk for Multi-Branch-Navigator
"""

import subprocess
import configparser
import time
from pathlib import Path
from datetime import datetime
import lib

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, Gio, GObject, GLib

from lib.alpm import AlpmBranch
from pyalpm import vercmp

#from memory_profiler import profile
#from memory_profiler import memory_usage
import gc

class GioConfig():
    # https://lazka.github.io/pgi-docs/Gio-2.0/classes/Settings.html ?
    KEY = 'GENERAL'
    inifile = Path.home() / f".config/manjaro/{lib.name()}.ini"

    def __init__(self):
        self.inifile.parent.mkdir(parents=True, exist_ok=True)
        self.config = configparser.ConfigParser()
        self.config.read(self.inifile)
        try:
            self.config[self.KEY]
        except KeyError:
            self.config[self.KEY] = {}

    def get_window_size(self):
        try:
            w, h = self.config['size'].get('w', 670), self.config['size'].get('h', 250)
            return int(w), int(h)
        except (KeyError, ValueError):
            return 670, 250

    def set_window_size(self, w: int, h: int):
        self.config['size'] = {}
        self.config['size']['w'] = str(w)
        self.config['size']['h'] = str(h)
        self.save()

    def get_fields(self):
        try:
            return self.config['FIELDS']
        except KeyError:
            return {'desc': True}

    def set_fields(self, fields: dict):
        self.config['FIELDS'] = {}
        for field in fields:
            self.config['FIELDS'][field] = '1'
        if not self.config['FIELDS']:
            self.config['FIELDS'] = {'desc': '1'}
        self.save()

    def save(self):
        with open(self.inifile, 'w') as configfile:
            self.config.write(configfile)


class PopupFields(Gtk.MenuButton):
    """ package field list """
    def __init__(self):
        super(PopupFields, self).__init__()
        self.set_property("tooltip-text", "search fields")
        self.field = 'name'

        self.menu = Gtk.Menu()
        self.set_popup(self.menu)

        branch = AlpmBranch('unstable')
        pkg = branch.pkg_get('pacman')
        fields = (a for a in sorted(dir(pkg)) if not '_' in a)
        group = None
        for field in fields:
            if field in ('md5sum', 'size', 'isize', 'files', 'deltas', 'backup', 'reason', 'sha256sum'):
                continue
            item = Gtk.RadioMenuItem(label=field)
            if not group:
                group = item
            else:
                item.set_property("group", group)
            if field == 'name':
                self.field = field
                item.set_active(True) # check
            item.show()
            item.connect('toggled', self.pop_action, field)
            self.menu.append(item)

    def pop_action(self, radiomenuitem, parameter):
        self.field = radiomenuitem.get_label()

class RightPopMenu(Gtk.Menu):
    """ popup menu for select columns in treeview """
    def __init__(self):
        super(RightPopMenu, self).__init__()
        fields_config = GioConfig().get_fields()
        self.fields = {}
        for i, item in enumerate(MainApp.FIELDS):
            if i < 3:
                continue
            menuitem = Gtk.CheckMenuItem(label=item)
            #menuitem.show()
            self.append(menuitem)
            is_active = fields_config.get(item, '')
            menuitem.set_active(is_active)
            if is_active:
                self.fields[item] = True
            menuitem.connect('toggled', self.pop_action, item)

    def pop_action(self, menuitem, parameter):
        if menuitem.get_active():
            self.fields[parameter] = True
        else:
            self.fields.pop(parameter, None)
        self.send_change(self.fields)

    def send_change(self, fields: dict):
        raise NotImplementedError()

class InfoDlg(Gtk.InfoBar):
    """ dialog informations """
    def __init__(self):
        super().__init__()
        self.message = ''
        self.pkg = None
        self.set_message_type(Gtk.MessageType.OTHER)
        self.set_show_close_button(True)
        self.set_default_response(0)
        self.set_revealed(True)
        self.connect("response", self.on_remove_info)
        self.info_label = Gtk.Label()
        self.info_label.set_line_wrap(True)
        vbox = Gtk.VBox(spacing=6)
        vbox.pack_start(self.info_label, expand=True, fill=True, padding=0)
        self.info_url = Gtk.Label()
        vbox.pack_start(self.info_url, expand=True, fill=True, padding=0)

        self.hbox = Gtk.HBox(spacing=6)
        self.pkgs = dict()
        for branchname in AlpmBranch.BRANCHES:
            data = {
                'contener': Gtk.Frame(label=branchname),
                'version' : Gtk.Label(label=''),
                'date' : Gtk.Label(label='')
            }
            self.pkgs[branchname] = data
            xbox = Gtk.VBox(spacing=4)
            xbox.pack_start(data['version'], expand=False, fill=True, padding=0)
            xbox.pack_end(data['date'], expand=False, fill=True, padding=0)
            data['contener'].add(xbox)
            self.hbox.pack_start(data['contener'], expand=True, fill=True, padding=0)
        vbox.pack_end(self.hbox, expand=True, fill=True, padding=0)
        self.get_content_area().pack_start(vbox, expand=True, fill=True, padding=0)
        self.hide()

    def on_remove_info(self, panel: Gtk.InfoBar, btnid: str):
        panel.hide()

    def set_message(self, msg: str, typed=None):
        """ type: Gtk.MessageType.WARNING ERROR, INFO, OTHER  """
        if not typed:
            typed = Gtk.MessageType.OTHER
        self.set_message_type(typed)
        self.info_label.set_markup(msg)
        self.show()
        self.info_url.hide()
        self.hbox.hide()

    def set_pkg(self, pkg_name):
        """ make pacman -Qi """
        # return no bug memory
        '''
        powerpill 40 Mo not in arch -> 2eme pass = 0 !!?
        pacui(community) 0Mo not in arch
        update-notifier(community) 0Mo not in arch
        trizen(community) 0Mo not in arch
        python3-xcpf 0Mo not in arch
        pkgfile(extra) 20Mo in all  > 2eme pass = 0 !!?
        pacman(core) 1Mo in all
        expac(extra) 0 Mo in all
        pacredir(community) 0Mo in all
        '''
        desc = ''
        url = ''

        for branchname in AlpmBranch.BRANCHES:
            branch = AlpmBranch(branchname)
            #mem_usage = memory_usage(-1, interval=.02, timeout=0.1, max_usage=True)
            #print(pkg_name , mem_usage)
            pkg = branch.pkg_get(pkg_name)
            #mem_usage = memory_usage(-1, interval=.02, timeout=0.1, max_usage=True)
            #print(pkg_name, mem_usage,"\n")
            if pkg:
                self.pkgs[branchname]['version'].set_label(pkg.version)
                self.pkgs[branchname]['date'].set_label(datetime.fromtimestamp(pkg.builddate).strftime('%Y-%m-%d'))
                self.pkgs[branchname]['contener'].show()
                desc = pkg.desc
                url = pkg.url
            else:
                self.pkgs[branchname]['contener'].hide()
            branch = None
            pkg = None
            gc.collect()
        self.set_message(f"<b>{pkg_name}</b>\n<i>{desc}</i>\n", Gtk.MessageType.INFO)
        self.info_url.set_markup(f'<a href="{url}">{url}</a>\n')
        self.info_url.show()
        self.hbox.show()


class MainApp:
    """ main window """

    BRANCHES = AlpmBranch.BRANCHES
    FIELDS = ('name', 'version', 'ver', 'desc', 'builddate', 'url', 'db', 'packager')

    def __init__(self, title: str = 'pacman'):
        self.setup = GioConfig()
        self.window = None
        self.store = None
        self.create_store('')
        # Use fixed-height-mode to get better model load and display performance.
        #view = Gtk.TreeView(model=store, fixed_height_mode=True, headers_visible=True)
        self.treeview = Gtk.TreeView(model=self.store, headers_visible=True, fixed_height_mode=False)
        self.treeview.connect("button-press-event", self.on_tree_click)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn('package', cell, text=0)
        column.set_sort_order(Gtk.SortType.DESCENDING)
        column.set_sort_column_id(0)
        column.set_resizable(True)
        self.treeview.append_column(column)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn('version', cell, markup=1)
        column.set_resizable(True)
        self.treeview.append_column(column)
        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn('version', cell, markup=2)
        column.set_sort_order(Gtk.SortType.DESCENDING)
        column.set_sort_column_id(2)
        column.set_resizable(True)
        column.set_visible(False)
        self.treeview.append_column(column)

        for i, item in enumerate(self.FIELDS):
            if i < 3:
                continue
            cell = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn(item, cell, text=i)
            column.set_sort_column_id(i)
            self.treeview.append_column(column)

        scrolled = Gtk.ScrolledWindow()
        scrolled.add(self.treeview)

        self.window = Gtk.Window(title='Multi-Branch-Navigator')
        self.window.connect('delete-event', self.on_close)
        self.window.connect('destroy', Gtk.main_quit)
        self.window.set_border_width(4)
        window_size = self.setup.get_window_size()
        self.window.set_default_size(window_size[0], window_size[1])
        self.window.set_position(Gtk.WindowPosition.CENTER_ALWAYS)

        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = self.window.get_title()
        self.window.set_titlebar(hb)

        self.update_process = None
        self.button_update = Gtk.Button(label='🗘')
        self.button_update.set_property("tooltip-text", "Update")
        self.button_update.connect("clicked", self.on_update)
        hb.pack_end(self.button_update)

        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        button = Gtk.Button()
        icon = Gio.ThemedIcon(name="go-home")
        button.set_property("tooltip-text", "clear list")
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        button.add(image)
        button.connect("clicked", self.on_branch_combo_changed)
        box.add(button)

        self.branch_combo = Gtk.ComboBoxText()
        self.branch_combo.set_entry_text_column(0)
        for branch in self.BRANCHES:
            self.branch_combo.append_text(branch)
        self.branch_combo.set_active(int(self.setup.config['GENERAL'].get('branch', '1')))
        self.branch_combo.connect("changed", self.on_branch_combo_changed)
        self.branch_combo.set_property("tooltip-text", "branch")
        box.pack_start(self.branch_combo, False, False, True)

        self.branch_combo_two = Gtk.ComboBoxText()
        self.branch_combo_two.set_entry_text_column(0)
        self.branch_combo_two.append_text('')
        for branch in self.BRANCHES:
            self.branch_combo_two.append_text(branch)
        self.branch_combo_two.set_active(0)
        self.branch_combo_two.connect("changed", self.on_branch_combo_changed)
        self.branch_combo_two.set_property("tooltip-text", "branch")
        box.pack_start(self.branch_combo_two, False, False, True)

        self.entry = Gtk.Entry()
        self.entry.set_property("tooltip-text", "to search")
        self.entry.connect("activate", self.on_search)
        self.entry.set_text(self.setup.config['GENERAL'].get('entry', ''))
        box.pack_start(self.entry, expand=True, fill=False, padding=0)

        self.fields = PopupFields()
        box.pack_start(self.fields, expand=False, fill=False, padding=0)

        button = Gtk.Button()
        icon = Gio.ThemedIcon(name="search")
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        button.add(image)
        button.set_property("tooltip-text", "Search")
        button.connect("clicked", self.on_search)
        box.pack_start(button, expand=True, fill=False, padding=0)

        hb.pack_start(box)

        vbox = Gtk.VBox(spacing=4)
        self.info = InfoDlg()
        vbox.pack_end(self.info, expand=False, fill=False, padding=0)
        vbox.pack_start(scrolled, expand=True, fill=True, padding=0)
        self.window.add(vbox)

        self.rmenu = RightPopMenu()
        self.rmenu.send_change = self.on_change_colums
        self.rmenu.send_change(self.rmenu.fields) # update from .config
        self.window.show_all()
        self.info.hide()

    def main(self):
        try:
            self.window.set_icon_from_file(f"/usr/share/{lib.PKGNAME}/{lib.name()}.svg")
        except gi.repository.GLib.Error:
            pass
        return Gtk.main()

    def set_title(self):
        """ display count packages after search """
        if self.window and self.store:
            nb_pkg = f" {len(self.store)}" if len(self.store)>0 else ''
            self.window.set_title(f"Multi-Branch-Navigator {nb_pkg}")

    def on_close(self, widget, event=None, data=None):
        """ save pos/setups before close app """
        self.setup.config['GENERAL']['entry'] = self.entry.get_text().strip()
        self.setup.config['GENERAL']['branch'] = str(self.branch_combo.get_active())
        self.setup.set_window_size(self.window.get_size().width, self.window.get_size().height)
        self.setup.set_fields(self.rmenu.fields)
        Gtk.main_quit()

    def create_store(self, branchname: str = 'unstable'):
        """ set datas in store """

        def set_versions(v1, v2, compare):
            if compare > 0:
                return v1, f'<span foreground="red">{v2}</span>'
            elif compare < 0:
                return f'<span foreground="red">{v1}</span>', v2
            return v1, v2

        if self.store:
            self.store.clear()
        else:
            self.store = Gtk.ListStore(str, str, str, str, str, str, str, str, int)
        self.set_title()
        if not branchname:
            return self.store

        to_find = self.entry.get_text().strip()
        print(to_find)
        branchname_two = self.branch_combo_two.get_active_text()

        branch = AlpmBranch(branchname)
        if self.fields.field == 'name':
            pkgs = branch.pkg_search(to_find)
        else:
            pkgs = branch.pkg_search_by_field(to_find, self.fields.field)
        if branchname_two:
            branch = AlpmBranch(branchname_two)
        for pkg in pkgs:
            version_one = pkg.version
            version_two = ''
            compare = 99
            if branchname_two and branchname_two != branchname:
                pkgb = branch.pkg_get(pkg.name)
                if pkgb:
                    version_two = pkgb.version
                    compare = vercmp(pkg.version, version_two)
                    version_one, version_two = set_versions(version_one, version_two, compare)
                    pkgb = None
            if not branchname_two or (branchname_two and compare != 0):
                self.store.append([
                    pkg.name,
                    version_one,
                    version_two,
                    pkg.desc,
                    time.strftime('%Y-%m-%d', time.localtime(pkg.builddate)),
                    pkg.url,
                    pkg.db.name,
                    pkg.packager.split('<')[0],
                    compare])
        self.set_title()
        return self.store

    def on_branch_combo_changed(self, widget):
        """ user change branch """
        if self.branch_combo_two.get_active() == self.branch_combo.get_active() + 1:
            self.branch_combo_two.set_active(0)
            return
        self.create_store('')
        gc.collect()
        self.treeview.get_column(2).set_visible(self.branch_combo_two.get_active() > 0)
        if self.branch_combo_two.get_active() == 0:
            for i in range(1, 3):
                self.treeview.get_column(i).set_title('version')
        else:
            self.treeview.get_column(1).set_title(self.branch_combo.get_active_text())
            self.treeview.get_column(2).set_title(self.branch_combo_two.get_active_text())

    def on_change_colums(self, fields):
        """ show/hide tree colums from menu and .config """
        for i, item in enumerate(MainApp.FIELDS):
            if i < 3:
                continue
            self.treeview.get_column(i).set_visible(fields.get(item, ''))

    def on_search(self, btn):
        text = self.branch_combo.get_active_text()
        print(text)
        watch_cursor = Gdk.Cursor(Gdk.CursorType.WATCH)
        self.treeview.get_window().set_cursor(watch_cursor)
        try:
            self.create_store(text)
            gc.collect()
        finally:
            self.treeview.get_window().set_cursor(None)

    def on_tree_click(self, treeview: Gtk.TreeView, event):
        """ double clic on tree view """
        if event.button == 1 and int(event.type == 5):  # Gdk.EventType.GDK_2BUTTON_PRESS:
            model, itertree = treeview.get_selection().get_selected()
            if not itertree:
                return
            pkg_name = model.get(itertree, 0)[0]
            if pkg_name:
                self.info.set_pkg(pkg_name)
            gc.collect()
        """ right mouse click """
        if event.button == 3:
            self.rmenu.show_all()
            self.rmenu.popup(None, None, None, None, event.button, event.time)


    def on_update(self, btn):
        """ update 4 branches """
        self.update_process = subprocess.Popen([
            'pkexec',
            f"/usr/share/{lib.PKGNAME}/__main__.py",
            'update',
            '--gui'
        ], stdout=None, stderr=None)
        self.button_update.set_sensitive(False)
        GLib.timeout_add(1000*20, self.on_timer_update_check) # check result in 20 seconds

    def on_timer_update_check(self):
        if self.update_process:
            if self.update_process.poll() == None:
                GLib.timeout_add(1000*20, self.on_timer_update_check)
                self.button_update.set_sensitive(False)
                return
            else:
                ret_code = self.update_process.poll()
                print('End update, returncode: ', ret_code)
                if ret_code > 0:
                    self.info.set_message(f"update ERROR: {ret_code}", Gtk.MessageType.WARNING)
                else:
                    self.info.set_message(f"databases updated", Gtk.MessageType.OTHER)
                self.update_process = None
        self.button_update.set_sensitive(True)


if __name__ == "__main__":
    print(f"{lib.name()} {lib.version_str()} {lib.DESCRIPTION} ")
    app = MainApp(title=lib.DESCRIPTION)
    #mem_usage = memory_usage(-1, interval=.02, timeout=0.1)
    #print('begin',mem_usage,"\n")
    app.main()
    #mem_usage = memory_usage(-1, interval=.02, timeout=0.1)
    #print('begin',mem_usage,"\n")
    #nb = gc.collect()
    #mem_usage = memory_usage(-1, interval=.02, timeout=0.1)
    #print('gc released=',nb,mem_usage,"\n")



"""
    Uses Alpm for request pacman databases
"""

import os
import subprocess
import re
from pathlib import Path
import tarfile
import json
import urllib.request
from datetime import datetime
from typing import List, Generator

import pyalpm
from pycman import config
import lib

from .exceptions import BranchNotFoundError

class AlpmBranch:
    """Multi branches APPM functions"""

    #BRANCHES = ('archlinux', 'unstable', 'testing', 'stable-staging', 'stable')
    BRANCHES = ('archlinux', 'unstable', 'testing', 'stable')
    ROOT = lib.read_config_root()
    COLORS = {
        'archlinux': '\33[34m',
        'unstable': '\33[31m',
        'testing': '\33[33m',
        'stable-staging': '\33[35m',
        'stable': '\33[32m',
        'none': '\33[0m',
        'i': '\33[3m'
    }
    handles = {}    # alpm handles

    def __new__(cls, branch='unstable'):
        # create 4 handles singleton
        if not cls.handles:
            for branchname in cls.BRANCHES:
                try:
                    cls.handles[branchname] = config.init_with_config(f"{cls.ROOT}/{branchname}/etc/pacman.conf")
                except FileNotFoundError:
                    if branch == branchname:
                        raise BranchNotFoundError(f"local repo \"{branchname}\" not found ! (use command \"update\")", branchname)
        return super().__new__(cls)

    def __init__(self, branch: str = 'unstable'):
        #self.handle = None
        self._branch = 'unstable'
        self.branch = branch

    @property
    def handle(self):
        """ return curent branch alpm handle """
        return self.handles[self._branch]

    @property
    def branch(self):
        """return curent branch"""
        return self._branch

    @branch.setter
    def branch(self, value: str):
        """set branch change pacman.conf"""
        if not value in self.BRANCHES:
            raise ValueError(f"{value} is not a valid branch {self.BRANCHES}")
        self._branch = value

    @classmethod
    def color(cls, color: str = 'none'):
        """one color by branch"""
        return cls.COLORS.get(color, 'none')

    def pkg_get(self, pkgname: str) -> object:
        """find one package in db"""
        for db_repo in self.get_repos():
            pkg = db_repo.get_pkg(pkgname)
            if pkg:
                return pkg
        return None

    def pkg_search(self, to_search: str) -> Generator:
        """search in databases in names and description"""
        for db_repo in self.get_repos():
            pkgs = db_repo.search(to_search)
            for pkg in pkgs:
                yield pkg

    def pkg_search_by_field(self, to_search: str, field: str = 'name') ->Generator:
        """ search in database by one package field """
        regex = re.compile(to_search)
        matchs = None
        for db_repo in self.get_repos():
            pkgs = db_repo.search('')
            for pkg in pkgs:
                if field == "builddate":
                    try:
                        value = datetime.fromtimestamp(pkg.builddate).strftime('%Y-%m-%d')
                    except:
                        value = ''
                else:
                    try:
                        value = getattr(pkg, field)
                    except AttributeError:
                        raise AttributeError(f"Attribut not exists for a package !\n\n{'  '.join((a for a in dir(pkg) if not '_' in a))}")
                matchs = regex.search(str(value))
                if matchs:
                    yield pkg

    def pkg_search_in_name(self, pkgname: str) -> Generator:
        """search in databases by only names"""
        regex = re.compile(pkgname)
        for db_repo in self.get_repos():
            pkgs = db_repo.search('')
            for pkg in pkgs:
                matchs = regex.match(pkg.name)
                if matchs:
                    yield pkg

    def get_repos(self) -> Generator:
        """list repos in curent branch"""
        for db_repo in self.handle.get_syncdbs():
            yield db_repo

    @staticmethod
    def add_sudo(args: list) -> list:
        if Path(AlpmBranch.ROOT).stat().st_uid == 0:
            args.insert(0, 'sudo')
        return args

    @classmethod
    def create_branch(cls, branch: str = 'unstable'):
        """create database structure from config files in /etc/"""
        if os.geteuid() != 0:
            raise PermissionError("use sudo for create local databases")
        Path(f"{cls.ROOT}/{branch}/etc").mkdir(parents=True, exist_ok=True)
        Path(f"{cls.ROOT}/hooks").mkdir(exist_ok=True)
        Path(f"{cls.ROOT}/{branch}/lib/pacman/sync").mkdir(parents=True, exist_ok=True)
        # write pacman.conf
        Path(f"{cls.ROOT}/{branch}/etc/pacman.conf").touch(644, True)
        if branch == 'archlinux':
            txt = Path(f"/etc/{lib.PKGNAME}/archlinux.conf").read_text()
        else:
            txt = Path(f"/etc/{lib.PKGNAME}/manjaro.conf").read_text()
            #change branch : .../packages/unstable/$repo/...
            txt = re.sub('/(unstable|stable|testing)/', f"/{branch}/", txt)
        txt = re.sub('\nRootDir.*', f"\nRootDir = {cls.ROOT}/{branch}", txt)
        txt = re.sub('\nDBPath.*', f"\nDBPath = {cls.ROOT}/{branch}/lib/pacman", txt)
        if re.search('\nHookDir', txt):
            txt = re.sub('\nHookDir.*', f"\nHookDir = {cls.ROOT}/hooks", txt)
        else:
            txt = txt.replace('[options]', f"[options]\nHookDir = {cls.ROOT}/hooks", 1)
        Path(f"{cls.ROOT}/{branch}/etc/pacman.conf").write_text(txt)
        Path(f"{cls.ROOT}/{branch}/etc/pacman.conf").chmod(mode=0o644)
        with open(f"{cls.ROOT}/hooks/00-no.hook", 'w') as writer:
            writer.write(
                "[Trigger]\nOperation = Install\nType = File\nTarget = *\n" + \
                "[Action]\nDescription = no\nWhen = PreTransaction\nExec = /usr/bin/false\nAbortOnFail\n"
            )

    @staticmethod
    def set_server(url: str) -> bool:
        """change server in pacman.conf"""
        '''
            #TODO for python 3.11
        s    use import tomllib
        '''
        #Server = https://mirror.netzspielplatz.de/manjaro/packages/stable/$repo/$arch

        if os.geteuid() != 0:
            raise PermissionError("use sudo for create local databases")

        conf_file = Path(f"/etc/{lib.PKGNAME}/manjaro.conf")
        txt = conf_file.read_text()
        m = re.search('\nServer.*', txt)
        result = m.group(0) if m else ""
        if result:
            result = result.split("=")[-1].strip()
            txt = txt.replace(result, f"{url}unstable/$repo/$arch")
            tmp_file = f"{lib.TMP_DIR}/pacman.conf"
            conf_file.write_text(txt)
            print(conf_file, "Updated")
        else:
            raise ValueError(conf_file + "without Server ?")

    def do_refresh(self, force: bool = False, use_files: bool = False) -> None:
        "Sync databases like pacman -Sy"
        if force:
            force = 'y'
        else:
            force = ''
        argsp = [
            'sudo', 'pacman', f"-Sy{force}",
            '--logfile', f"{self.ROOT}/{self.branch}/pacman.log",
            '--config', f"{self.ROOT}/{self.branch}/etc/pacman.conf"
        ]
        try:
            subprocess.check_call(argsp)
            # load databases.files
            if use_files:
                argsp[2] = "-Fy"
                subprocess.check_call(argsp)
        except KeyboardInterrupt:
            raise PermissionError('Update aborded, Interrupted by the user')

    def do_update(self, force: bool = False) -> bool:
        """ as self.do_refresh() but use only alpm
            not use files
            Want exeception for gui
        """
        _last_dl_filename = ''

        def dl_cb(filename, tx, total):
            nonlocal _last_dl_filename
            if filename != _last_dl_filename:
                _last_dl_filename = filename
                print(f"\tdownload {filename} ...")

        self.handle.dlcb = dl_cb
        try:
            transaction = self.handle.init_transaction(force=force)
        except pyalpm.error as err:
            raise err
        try:
            for db_repo in self.handle.get_syncdbs():
                _last_dl_filename = ''
                db_repo.update(force=force)
        except pyalpm.error as err:
            raise err
        finally:
            transaction.release()
            self.handle.dlcb = None
        return True

    @classmethod
    def read_files(cls, branch, pkg: object) -> list:
        """extract file list from a package"""
        result = []
        if not os.path.exists(f"{cls.ROOT}/{branch}/lib/pacman/sync/{pkg.db.name}.files"):
            return result
        with tarfile.open(f"{cls.ROOT}/{branch}/lib/pacman/sync/{pkg.db.name}.files", 'r') as tar:
            tar.extract(f"{pkg.name}-{pkg.version}/files", lib.TMP_DIR)
            with open(f"{lib.TMP_DIR}/{pkg.name}-{pkg.version}/files") as ffiles:
                for line in ffiles:
                    line = line.strip()
                    if not line.endswith('/') and not line.startswith('%'):
                        result.append(line)
            os.remove(f"{lib.TMP_DIR}/{pkg.name}-{pkg.version}/files")
            os.rmdir(f"{lib.TMP_DIR}/{pkg.name}-{pkg.version}")
        return result

    def find_files(self, to_find: str, only_one: bool = False) -> List:
        """find in files list"""
        regex = re.compile(to_find)
        result = []
        for db_repo in self.get_repos():
            pkg = {'name':'', 'version':'', 'files':[], 'db':db_repo.name}
            try:
                with tarfile.open(f"{self.ROOT}/{self.branch}/lib/pacman/sync/{db_repo.name}.files", 'r') as tar:
                    tar.extractall(lib.TMP_DIR)
            except FileNotFoundError:
                raise FileNotFoundError(f"run update --files before !")

            for item in Path(lib.TMP_DIR).glob('*/files'):
                with open(item, 'r') as f_files:
                    pkg = {'name':'', 'version':'', 'files':[], 'db':db_repo.name}
                    for line in f_files:
                        if regex.search(line):
                            pkg['files'].append(line.rstrip())
                    if pkg['files']:
                        with open(Path(item).parent / 'desc', 'r') as f_desc:
                            for line in f_desc:
                                line = line.rstrip()
                                if not line:
                                    continue
                                if line == '%NAME%':
                                    pkg['name'] = next(f_desc).strip()
                                if line == '%VERSION%':
                                    pkg['version'] = next(f_desc).strip()
                                    break
                        result.append(pkg)
                        title = f"{pkg['name']} {pkg['version']}"
                        print(f"{self.color(self.branch)}::{self.color('none')} {title:68} ({pkg['db']})")
                        for fpkg in pkg['files']:
                            print(f"\t{fpkg}")
                        if only_one:
                            subprocess.run(['rm', '-rf', lib.TMP_DIR])
                            os.mkdir(lib.TMP_DIR)
                            return result
            subprocess.run(['rm', '-rf', lib.TMP_DIR])
            os.mkdir(lib.TMP_DIR)
        return result

    @staticmethod
    def aur_query(pkgname: str) -> None:
        """get aur pkg infos"""

        url = f"https://aur.archlinux.org/rpc/?v=5&type=info&arg[]={pkgname}"
        with urllib.request.urlopen(url) as q_aur:
            req = q_aur.read().decode('utf-8')
        data = json.loads(req)
        if data['resultcount'] < 1 or data['type'] == "error":
            return None

        pkg = data['results'][0]
        print(f"{'AUR':15}: ------")
        print(f"{'Name':15}: {pkg['Name']}")
        print(f"{'Version':15}: {pkg['Version']}")
        print(f"{'Date ':15}: {datetime.fromtimestamp(pkg['LastModified']).strftime('%a %d %b %Y %X %Z')}")
        print(f"{'Url ':15}: {'https://aur.archlinux.org/packages/'+pkgname}") # or URLPath

    @staticmethod
    def aur_infos(pkgs, verbose: bool = False):
        """
        390 packages in aur 2019-04-05 , 4000 requests per day per IP.
        URI maximum length limit of 4443 bytes
        200 packages MAXI in multi info/search
        """
        # tests
        # 200 pkgs => 4100 len url !
        # 150 => 3200 :)
        # 140 => 3000
        MAX_PKGS = 140
        URL = f"https://aur.archlinux.org/rpc/?v=5&type=info"

        if not pkgs:
            return {}

        result = {}
        urls = []   # only MAX_PKGS by url
        url = URL
        i = 0
        for i, pkg in enumerate(pkgs):
            if i % MAX_PKGS == 0:
                urls.append(url)
                url = URL
            url += f"&arg[]={pkg}"
        if url != URL:
            urls.append(url)
            #if verbose:
            #    print("url", url)

        if len(urls) > 2:
            print(len(urls)-1, "request(s) aur")
        for url in urls:
            if url == URL:
                continue
            if len(url) > 4440:
                raise IndexError(f"URI maximum length limit of 4443 bytes : here {len(url)}")

            with urllib.request.urlopen(url) as q_aur:
                req = q_aur.read().decode('utf-8')
            data = json.loads(req)
            if data['resultcount'] < 1 or data['type'] == "error":
                print("\n\nAUR ERROR !!!\n", data, "\nURL:", url)

            for pkg in data['results']:
                result[pkg['Name']] = pkg['Version']

        #print(f"\nAUR packages ({len(result)}):\n", result)
        return result



"""
        search in databases
"""
import re
import time
from datetime import datetime
from datetime import timedelta
from lib.actions import Action
from lib.alpm import AlpmBranch
from lib.displays import DisplayPkg


class ActionSearch(Action):

    @staticmethod
    def usage():
        return """
            search regex --unstable [--desc] [--quiet or -q] [--packager] [--dep] : search in one branch, by default only in pkgname
                --dep [filter]: dependence start by .. , if no "filter" use "regex"
                --packager [filter] : search regex by packager , if no "filter" use "regex"
                --quiet : no show description
                --desc : search in description and package name
                --field field_name : search 'regex' in one field (name, url, buildate,...)
                --format '%n %v %e %d %G %b %p %r %u %P %D' (as expac)
            """

    def test(self):
        super().test()
        if not self.params.branch:
            self.params.branch = 'unstable'
        return True

    def run(self, to_find: str = ''):
        """ command : seach packages by name or description """
        super().run()
        self.msg(f"Branch: {AlpmBranch.color(self.params.branch)}{self.params.branch}{AlpmBranch.color('none')}")
        in_desc = self.params.option('desc')
        by_dep = self.params.option('dep')
        by_date = self.params.option('date')
        by_packager = self.params.option('packager')
        format_model = self.params.option('format')
        by_field = self.params.option('field')
        quiet = self.params.option('quiet', 'q')
        if not to_find:
            try:
                to_find = self.params.values.pop(0)
            except IndexError:
                to_find = ''
        if by_dep:
            if isinstance(by_dep, bool):
                by_dep = to_find
                to_find = ''
            in_desc = True
        if by_packager:
            if isinstance(by_packager, bool):
                by_packager = re.compile(f"<.*{to_find}")
                to_find = ''
            else:
                by_packager = re.compile(f"<.*{by_packager}")
            in_desc = True
        if by_date:
            to_date_reverse = False
            if to_find.endswith('-'):
                to_find = to_find[:-1]
            if to_find.endswith('+'):
                to_find = to_find[:-1]
                to_date_reverse = True
            if to_find.startswith('-'):
                to_date_reverse = True
                by_date = time.mktime((datetime.today() + timedelta(days=int(to_find))).timetuple())
            else:
                by_date = time.mktime(datetime.strptime(to_find, '%Y-%m-%d').timetuple())
            in_desc = True
            to_find = ''

        branch = AlpmBranch(self.params.branch)
        if not format_model:
            if quiet:
                format_model = "{$d%n$n %v:72} (%r)"
            else:
                format_model = "{$d%n$n %v:72} (%r)\n\t$i%d$n"
        display = DisplayPkg(format_model, self.params.branch)
        found = False
        if in_desc:
            for pkg in branch.pkg_search(to_find):
                if by_dep and not [x for x in pkg.depends if x.startswith(by_dep)]:
                    continue
                if by_packager and not by_packager.search(pkg.packager):
                    continue
                if by_date:
                    if to_date_reverse:
                        if by_date > pkg.builddate:
                            continue
                    else:
                        if by_date < pkg.builddate:
                            continue
                found = True
                print(display.string(pkg))
                if by_date:
                    print(f"\t{branch.color('i')}{time.strftime('%a %d %b %Y', time.localtime(pkg.builddate))}{branch.color('none')}")
        else:
            if isinstance(by_field, bool):
                by_field = 'name'
            for pkg in branch.pkg_search_by_field(to_find, by_field):
                found = True
                print(display.string(pkg))
        if not found:
            raise UserWarning(f"Nothing found ! {to_find}")



"""
    Action pactree
"""

import subprocess
from lib.actions import Action
from lib.alpm import AlpmBranch


class ActionPactree(Action):

    @staticmethod
    def usage():
        return "pactree : as original command, only --sync"

    def test(self):
        super().test()
        if not self.params.values:
            raise ValueError(f"Nothing to find !")
        return True

    def run(self, pkg_name: str = ''):
        """ main """
        super().run()
        use_reverse = self.params.option('reverse', 'r')
        use_list = self.params.option('linear', 'l')
        use_depth = self.params.option('depth', 'd')
        if not pkg_name:
            pkg = self.params.values.pop(0)
        else:
            pkg = pkg_name
            use_list = True
        result = ''

        for branch_name in AlpmBranch.BRANCHES:
            if self.params.branch and branch_name != self.params.branch:
                continue
            args = ['pactree', pkg, '-s', '-b', f"{AlpmBranch.ROOT}/{branch_name}/lib/pacman"]
            if use_reverse:
                args.append('-r')
            if use_list:
                args.append('-l')
                args.append('-u')
            if use_depth and use_depth.isdigit():
                args.append(f"-d{use_depth}")
            if not pkg_name:
                self.msg(f"Branch: {AlpmBranch.color(branch_name)}{branch_name}{AlpmBranch.color('none')}")

            if pkg_name:
                #capture call output in result
                process, error = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
                if not error:
                    result += process.decode()
            else:
                subprocess.call(args)

        if pkg_name:
            return result.split()



"""
        Action diff
"""
from lib.actions import Action
from lib.alpm import AlpmBranch
from lib.displays import DisplayPkg
from lib.exceptions import BranchError


class ActionDiff(Action):

    @staticmethod
    def usage():
        return """
            diff --testing --stable [--rm | --add] : compare two branches, can show only one column
                --rm [optional] : only next remove
                --add [optional] : only add in next branch
                --format '%n %v %e %d %G %b %p %r %u %P %D' only with --rm and --add
            """

    def test(self):
        super().test()
        if not self.params.branch:
            raise BranchError(f"Enter a valid branch {AlpmBranch.BRANCHES}")
        if not self.params.branch_two:
            raise BranchError(f"Enter a second valid branch {AlpmBranch.BRANCHES}")
        return True

    def run(self):
        """Diff package list for 2 branches"""
        super().run()
        branches = self.params.get_branches()
        a_branch = AlpmBranch(branches[0])
        items_a = {item.name for item in a_branch.pkg_search('')}
        b_branch = AlpmBranch(branches[1])
        items_b = {item.name for item in b_branch.pkg_search('')}
        only_in_b = items_b - items_a
        show_add = self.params.option('add')
        show_rm = self.params.option('rm')
        format_model = self.params.option('format')
        if not format_model:
            display = DisplayPkg("%n", 'none')
        nb_packages = 0
        ptitle = f"{branches[0]} ({len(items_a - items_b)})"
        title = f":: {AlpmBranch.color(branches[0])}{ptitle:40}{AlpmBranch.color('none')} / {AlpmBranch.color(branches[1])}{branches[1]} ({len(only_in_b)}){AlpmBranch.color('none')}"

        if show_rm:
            title = f":: removed in {AlpmBranch.color(b_branch.branch)}{b_branch.branch} ({len(only_in_b)}){AlpmBranch.color('none')}"
            print(title, end="\n\n")
            if format_model:
                display = DisplayPkg(format_model, b_branch.branch)
            for item in sorted(only_in_b):  # only not in A
                nb_packages += 1
                print(display.string(b_branch.pkg_get(item)))
            if nb_packages > 40:
                print(f"\n{title}\n")
            return
        if show_add:
            only_in_a = items_a - items_b
            title = f":: news in {AlpmBranch.color(a_branch.branch)}{a_branch.branch} ({len(only_in_a)}){AlpmBranch.color('none')}"
            print(title, end="\n\n")
            if format_model:
                display = DisplayPkg(format_model, a_branch.branch)
            for item in sorted(only_in_a):  # only not in B
                nb_packages += 1
                print(display.string(a_branch.pkg_get(item)))
            if nb_packages > 40:
                print(f"\n{title}")
            return

        # default show 2 colomns
        print(f"\n{title}")
        for item in sorted(items_a ^ items_b):  # all diffs
            nb_packages += 1
            prefix = ''
            if item in only_in_b:
                prefix = f"{prefix:46}"
            print(f"{prefix}{item}")
        if nb_packages > 40:
            print(f"\n{title}\n{nb_packages} packages")
